import type { FC, ReactNode } from 'react';
import React, { useCallback, useState, useEffect } from 'react';
import ReactDOM from 'react-dom';
import { styled } from '@storybook/theming';
import { global } from '@storybook/global';

import type { Config as ReactPopperTooltipConfig, PopperOptions } from 'react-popper-tooltip';
import { usePopperTooltip } from 'react-popper-tooltip';
import { Tooltip } from './Tooltip';

const { document } = global;

// A target that doesn't speak popper
const TargetContainer = styled.div<{ trigger: ReactPopperTooltipConfig['trigger'] }>`
  display: inline-block;
  cursor: ${(props) =>
    props.trigger === 'hover' || props.trigger.includes('hover') ? 'default' : 'pointer'};
`;

const TargetSvgContainer = styled.g<{ trigger: ReactPopperTooltipConfig['trigger'] }>`
  cursor: ${(props) =>
    props.trigger === 'hover' || props.trigger.includes('hover') ? 'default' : 'pointer'};
`;

interface WithHideFn {
  onHide: () => void;
}

export interface WithTooltipPureProps
  extends Omit<ReactPopperTooltipConfig, 'closeOnOutsideClick'>,
    PopperOptions {
  svg?: boolean;
  withArrows?: boolean;
  hasChrome?: boolean;
  tooltip: ReactNode | ((p: WithHideFn) => ReactNode);
  children: ReactNode;
  onDoubleClick?: () => void;
  /**
   * @deprecated use `defaultVisible` property instead. This property will be removed in SB 8.0
   */
  tooltipShown?: boolean;
  /**
   * @deprecated use `closeOnOutsideClick` property instead. This property will be removed in SB 8.0
   */
  closeOnClick?: boolean;
  /**
   * @deprecated use `onVisibleChange` property instead. This property will be removed in SB 8.0
   */
  onVisibilityChange?: (visibility: boolean) => void | boolean;
  /**
   * If `true`, a click outside the trigger element closes the tooltip
   * @default false
   */
  closeOnOutsideClick?: boolean;
}

// Pure, does not bind to the body
const WithTooltipPure: FC<WithTooltipPureProps> = ({
  svg,
  trigger,
  closeOnOutsideClick,
  placement,
  hasChrome,
  withArrows,
  offset,
  tooltip,
  children,
  closeOnTriggerHidden,
  mutationObserverOptions,
  closeOnClick,
  tooltipShown,
  onVisibilityChange,
  defaultVisible,
  delayHide,
  visible,
  interactive,
  delayShow,
  modifiers,
  strategy,
  followCursor,
  onVisibleChange,
  ...props
}) => {
  const Container = svg ? TargetSvgContainer : TargetContainer;
  const {
    getArrowProps,
    getTooltipProps,
    setTooltipRef,
    setTriggerRef,
    visible: isVisible,
    state,
  } = usePopperTooltip(
    {
      trigger,
      placement,
      defaultVisible: defaultVisible ?? tooltipShown,
      delayHide,
      interactive,
      closeOnOutsideClick: closeOnOutsideClick ?? closeOnClick,
      closeOnTriggerHidden,
      onVisibleChange: (_isVisible) => {
        onVisibilityChange?.(_isVisible);
        onVisibleChange?.(_isVisible);
      },
      delayShow,
      followCursor,
      mutationObserverOptions,
      visible,
      offset,
    },
    {
      modifiers,
      strategy,
    }
  );

  const tooltipComponent = (
    <Tooltip
      placement={state?.placement}
      ref={setTooltipRef}
      hasChrome={hasChrome}
      arrowProps={getArrowProps()}
      withArrows={withArrows}
      {...getTooltipProps()}
    >
      {typeof tooltip === 'function' ? tooltip({ onHide: () => onVisibleChange(false) }) : tooltip}
    </Tooltip>
  );

  return (
    <>
      <Container trigger={trigger} ref={setTriggerRef as any} {...props}>
        {children}
      </Container>
      {isVisible && ReactDOM.createPortal(tooltipComponent, document.body)}
    </>
  );
};

WithTooltipPure.defaultProps = {
  svg: false,
  trigger: 'click',
  closeOnOutsideClick: false,
  placement: 'top',
  modifiers: [
    {
      name: 'preventOverflow',
      options: {
        padding: 8,
      },
    },
    {
      name: 'offset',
      options: {
        offset: [8, 8],
      },
    },
    {
      name: 'arrow',
      options: {
        padding: 8,
      },
    },
  ],
  hasChrome: true,
  defaultVisible: false,
};

const WithToolTipState: FC<
  Omit<WithTooltipPureProps, 'onVisibleChange'> & {
    startOpen?: boolean;
    onVisibleChange?: (visible: boolean) => void | boolean;
  }
> = ({ startOpen = false, onVisibleChange: onChange, ...rest }) => {
  const [tooltipShown, setTooltipShown] = useState(startOpen);
  const onVisibilityChange = useCallback(
    (visibility: boolean) => {
      if (onChange && onChange(visibility) === false) return;
      setTooltipShown(visibility);
    },
    [onChange]
  );

  useEffect(() => {
    const hide = () => onVisibilityChange(false);
    document.addEventListener('keydown', hide, false);

    // Find all iframes on the screen and bind to clicks inside them (waiting until the iframe is ready)
    const iframes: HTMLIFrameElement[] = Array.from(document.getElementsByTagName('iframe'));
    const unbinders: (() => void)[] = [];
    iframes.forEach((iframe) => {
      const bind = () => {
        try {
          if (iframe.contentWindow.document) {
            iframe.contentWindow.document.addEventListener('click', hide);
            unbinders.push(() => {
              try {
                iframe.contentWindow.document.removeEventListener('click', hide);
              } catch (e) {
                // logger.debug('Removing a click listener from iframe failed: ', e);
              }
            });
          }
        } catch (e) {
          // logger.debug('Adding a click listener to iframe failed: ', e);
        }
      };

      bind(); // I don't know how to find out if it's already loaded so I potentially will bind twice
      iframe.addEventListener('load', bind);
      unbinders.push(() => {
        iframe.removeEventListener('load', bind);
      });
    });

    return () => {
      document.removeEventListener('keydown', hide);
      unbinders.forEach((unbind) => {
        unbind();
      });
    };
  });

  return (
    <WithTooltipPure {...rest} defaultVisible={tooltipShown} onVisibleChange={onVisibilityChange} />
  );
};

export { WithTooltipPure, WithToolTipState, WithToolTipState as WithTooltip };
